package org.cebs.corep.xbrl.util;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.UTFDataFormatException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.commons.lang.StringUtils;
import org.cebs.corep.excelToXbrl.constants.XbrlConstantsInterface;
import org.cebs.corep.xbrl.exceptions.CorepTechnicalException;
import org.cebs.corep.xbrl.log.CorepLogger;
import org.cebs.corep.xbrl.log.LogAction;
import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.xml.sax.InputSource;

/**
 * This class provides methods to check and validate xbrl instance
 * documents on xml schema level. The validation of the linkbases
 * is not included.
 */
public class DocumentValidator implements XbrlConstantsInterface
{
  final private static String CLASS = "DocumentValidator";
  public static final int NOT_WELL_FORMED = 0;
  public static final int NOT_WELL_ENCODED = 1;
  public static final int WELL_FORMED_AND_WELL_ENCODED = 2;
  private static final String XSD_PATH = "de/bundesbank/bakisw/mapping/util/"; //Pfad der Schema-Datei
  private static final String XSD = "InternesXMLSchema.xsd"; //Dateiname des XML-Schemas

  /**
   * Transforms the JDom document into a byte array and checks
   * if it is wellformed and valid according the xsd files.
   * @param doc
   * @param cc
   * @return
   */
  public static String validateDocument(Document doc, CorepClientContext cc)
  {
    final String METHOD = "validateDocument";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    String errorMessage = null;

    // converts the JDOM object into a byte array
    byte[] docAsBytes = JdomConverter.convertDocumentToByteArray(doc, cc);

    // if wellformed than validation against the taxonomies
    if (isWellFormed(docAsBytes, cc) == WELL_FORMED_AND_WELL_ENCODED)
      errorMessage = isValidAccordingToSchema(docAsBytes, cc);

    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);

    // possible error message
    return errorMessage;
  }

  /**
   * Checks if the document is wellformed.
   * @param docAsBytes 
   * @param cc  
   * @return boolean 
   */
  public static int isWellFormed(byte[] docAsBytes, CorepClientContext cc)
  {
    final String METHOD = "isWellFormed";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    // sets the parser for validation
    SAXBuilder builder = new SAXBuilder("org.apache.xerces.parsers.SAXParser", false);

    try
    {
      InputSource input = new InputSource(new ByteArrayInputStream(docAsBytes));
      input.setEncoding("UTF-8");
      // checks if the document is wellformed
      builder.build(input);
    }
    catch (JDOMException e)
    {
      CorepLogger.log(LogAction.COREP_XBRL_TECHNICAL_WARNING, CLASS, METHOD, null, "Document not wellformed.", e, cc);
      return NOT_WELL_FORMED;
    }
    catch (UTFDataFormatException e)
    {
      CorepLogger.log(LogAction.COREP_XBRL_TECHNICAL_WARNING, CLASS, METHOD, null, "Document has not the UTF-8 encoding.", e, cc);
      return NOT_WELL_ENCODED;
    }
    catch (IOException e)
    {
      CorepLogger.log(LogAction.COREP_XBRL_TECHNICAL_ERROR, CLASS, METHOD, null, "I/O error while parsing the document.", e, cc);
      throw new CorepTechnicalException();
    }
    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);
    return WELL_FORMED_AND_WELL_ENCODED;
  }

  /**
   * Validates the instance document against the xsd schema files of the taxonomy set.
   * Returns an error message if the document is not valid.
   * @param  docAsBytes 
   * @param  cc  
   * @return String
  */
  private static String isValidAccordingToSchema(byte[] docAsBytes, CorepClientContext cc)
  {
    final String METHOD = "isValidAccordingToSchema";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    try
    {
      // sets the parser for validation
      SAXBuilder builder = new SAXBuilder("org.apache.xerces.parsers.SAXParser", true);

      builder.setFeature("http://apache.org/xml/features/validation/schema", true);

      File zipFile = new File(Thread.currentThread().getContextClassLoader().getResource(XbrlConstantsInterface.SCHEMA_DIRECTORY + XbrlConstantsInterface.TAXONOMY_ZIP).getFile());

      String urlZipFile = zipFile.toURL().toString();

      ZipFile zip = new ZipFile(StringUtils.replace(StringUtils.replace(zipFile.getPath(), "file:", ""), "%20", " "));
      Enumeration entries = zip.entries();
      String prefix = null;
      // read the taxonomies.zip file and extracts the schema files   
      String taxonomyNamepaces = getTaxonomyNamespaces(urlZipFile, entries);
      // all xsd files that the instance might use are set 
      builder.setProperty(
        "http://apache.org/xml/properties/schema/external-schemaLocation",
        XBRLI_NS_URL
          + getURLString(urlZipFile, "xbrl-instance-2003-12-31.xsd")
          + LINKBASE_NS_URL
          + getURLString(urlZipFile, "xbrl-linkbase-2003-12-31.xsd")
          + XLINK_NS_URL
          + getURLString(urlZipFile, "xlink-2003-12-31.xsd")
          + XBRLDI_NS_URL
          + getURLString(urlZipFile, "xbrldi-2005.xsd")
          + taxonomyNamepaces);

      InputSource input = new InputSource(new ByteArrayInputStream(docAsBytes));
      input.setEncoding("UTF-8");
      // validates the instance document according the xsd files
      builder.build(input);
    }
    catch (JDOMException e)
    {
      CorepLogger.log(LogAction.COREP_XBRL_TECHNICAL_WARNING, CLASS, METHOD, null, "Dokument nicht gltig", e, cc);
      return e.getMessage();
    }
    catch (IOException e)
    {
      CorepLogger.log(LogAction.COREP_XBRL_TECHNICAL_ERROR, CLASS, METHOD, null, "I/O Fehler beim Parsen des Dokumentes", e, cc);
      throw new CorepTechnicalException();
    }
    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);

    return null;
  }

  /**
   * Extract the namespaces of the taxonomies out of the zip file.
   * @param urlZipFile
   * @param entries
   * @return
   */
  private static String getTaxonomyNamespaces(String urlZipFile, Enumeration entries)
  {
    String prefix;
    String taxonomyNamepaces = null;
    while (entries.hasMoreElements())
    {
      ZipEntry entry = (ZipEntry)entries.nextElement();
      if (entry.isDirectory())
      {
        continue;
      }

      if ((entry.getName().startsWith("t-") || entry.getName().startsWith("d-") || entry.getName().startsWith("p-")) && entry.getName().endsWith("xsd"))
      {

        prefix = StringUtils.left(entry.getName(), StringUtils.indexOf(entry.getName(), "-2005"));

        // fills in the external-schemaLocation for the schema taxonomies
        if (prefix.length() == 4)
        {
          taxonomyNamepaces = taxonomyNamepaces + getNamespaceURL(prefix) + getURLString(urlZipFile, entry.getName());
        }
      }
    }
    return taxonomyNamepaces;
  }

  /**
   * Gets the url character string for a schema file.
   * @param schemaFile
   * @return
   */
  private static String getURLString(String urlZipFile, String schemaFile)
  {
    return " " + "jar:" + urlZipFile + "!/" + schemaFile + " ";
  }

  /**
   * Returns the complete URL of the namespace of the taxonomy.
   * @param taxonomyPrefix
   * @return
   */
  protected static String getNamespaceURL(String taxonomyPrefix)
  {
    return NS_URL + taxonomyPrefix + NS_DATE;
  }
}
