package org.cebs.corep.xbrl.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;

import org.cebs.corep.xbrl.exceptions.CorepTechnicalException;
import org.cebs.corep.xbrl.log.CorepLogger;
import org.cebs.corep.xbrl.log.LogAction;
import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.xml.sax.InputSource;

/**
 * Transforms a JDom document into a character string.
 */
public class JdomConverter
{
  private static final String CLASS = "JdomToStringConverter";

  /**
   * Transforms a JDOM tree in to a character string with UTF-8 encoding.
   * @param doc 	Document 
   * @param cc		CorepClientContext
   * @return String 
   */
  public static String convertDocumentToString(Document doc, CorepClientContext cc)
  {
    final String METHOD = "convertDocumentToString";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    // use Jdom formatter
    XMLOutputter outputter = new XMLOutputter();

    // get pretty format
    Format format = Format.getPrettyFormat();

    // set the encoding to "UTF-8"
    outputter.setFormat(format.setEncoding("UTF-8"));

    // transforms the xml output into a character string
    String xmlOutput = outputter.outputString(doc);

    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);

    return xmlOutput;
  }

  /**
   * Transforms the JDom tree into an array of bytes.
   * @param doc Document 
   * @return String 
   */
  public static byte[] convertDocumentToByteArray(Document doc, CorepClientContext cc)
  {
    final String METHOD = "convertDocumentToByteArray";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    byte[] docAsBytes = null;
    // transforms a JDOM tree in to a character string 
    String docAsString = convertDocumentToString(doc, cc);
    // converts the character string into an array of bytes
    docAsBytes = convertStringToByteArray(docAsString, cc);

    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);
    return docAsBytes;
  }

  /**
   * Transforms a character string into an array of bytes with utf-8 encoding.
   * @param doc 	Document 
   * @param cc		CorepClientContext
   * @return String 
   */
  private static byte[] convertStringToByteArray(String docAsString, CorepClientContext cc)
  {
    final String METHOD = "convertStringToByteArray";

    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    byte[] docAsBytes = null;

    try
    {
      // convert to bytes with encoding UTF-8
      docAsBytes = docAsString.getBytes("UTF-8");
    }
    catch (UnsupportedEncodingException e)
    {
      CorepLogger.log(LogAction.TECHNICAL_ERROR, CLASS, METHOD, null, "Umwandlung nach UTF-8 gescheitert", e, cc);
      throw new CorepTechnicalException();
    }

    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);

    return docAsBytes;
  }

  /**
  	 * Wandelt Zeichenkette in einen JDOMBaum. Nimmt keine Validierung
  	 * gegen Schema oder DTD vor. Fr die Validierung dient die Klasse DocumentValidator.
  	 * @param stream
  	 * @param cc
  	 * @return
  	 */
  public static Document convertStreamToDocument(InputStream stream, CorepClientContext cc)
  {
    final String METHOD = "convertStreamToDocument";
    CorepLogger.log(LogAction.METHOD_BEGIN, CLASS, METHOD, cc);

    // das Ergebnis
    Document doc = null;

    // waehle Parser, Validierung ausgeschaltet, da bereits erfolgt
    SAXBuilder builder = new SAXBuilder("org.apache.xerces.parsers.SAXParser", false);

    try
    {
      InputSource input = new InputSource(stream);
      input.setEncoding("UTF-8");
      doc = builder.build(input);
    }
    catch (JDOMException e)
    {
      CorepLogger.log(LogAction.TECHNICAL_ERROR, CLASS, METHOD, null, "Dokument nicht wohlgeformt", e, cc);
      throw new CorepTechnicalException();
    }
    catch (IOException e)
    {
      CorepLogger.log(LogAction.TECHNICAL_ERROR, CLASS, METHOD, null, "I/O Fehler beim Parsen des Dokumentes", e, cc);
      throw new CorepTechnicalException();
    }
    CorepLogger.log(LogAction.METHOD_END, CLASS, METHOD, cc);
    return doc;
  }
}
